#define PUSHBUTTON_HEIGHT 18  // at least 13 (default: 18)
#define PUSHBUTTON_WIDTH  19  // at least 16 (default: 19)
#define HAS_HSCROLLBAR 0     // 16 if has an horizontal scrollbar

#define ELEVATOR_COLOR "Cyan"
#define UPBITMAP_COLOR "Black"
/*
Colors available:
   Ash
   Black (_not_ available for ELEVATOR_COLOR)
   Blue
   Brown
   Cyan
   Green
   Mauve
   Orange
   Red
   Yellow
*/

* Some usefull API-Constants
		#define wm_paint				 15	
		#define WM_VSCROLL          0x0115
		#define GW_CHILD            5
		#define GW_HWNDNEXT         2
		#define GWL_WNDPROC         (-4)
		#define SB_LINEUP           0
		#define SB_LINEDOWN         1
		#define SB_PAGEUP           2
	   #define SB_PAGEDOWN         3
	   #define SB_TOP              6
	   #define SB_BOTTOM           7
		#define GWL_STYLE           (-16)
		#define CS_DBLCLKS          0x0008
		#define PATAND              0x00A000C9 // dest = pattern AND dest    
      #define PATOR               0x00FA0089 // dest = pattern OR dest          
      #define DSTINVERT           0x00550009 // dest = NOT dest   
		#define PATNOTAND           0x000A0329 // dest = pattern AND NOT dest   
      #define PATNOTOR				 0x00F50225 // dest = pattern OR NOT dest	                 

/*
Code by Marc Van den Berghen

The Gigagrid is a very ambitious project, trying to add a lot of functionality
to the grid. The styling and general behaviour of the class has been designed 
by Jean-Pierre Martel

Goals : ok	 - use JP's grid and implement the standard behaviour (scrolling without navigation of the current row)
	  	  ok	 - enable smooth scrolling even for indexed rows (if the user wish)
		  ok   - implement the wheel-mouse
        ok   - make it compatible with the seeker
        ok   - avoid the appearance of the gray-zone, if there are enough rows
better not         - rowselect = true but still be editable
        ok   - solve the bug with disappearing lines when no columnheadings
        ok	 - adding a proportional-property, clicking in the shaft moves the grid to this position
delayed      - adding a color-in-function-of-value property to the cells
delayed      - providing a background-capability (like alternating colored lines) for grid and header
  next       - changing the click-behaviour of the buttons (click and hold)
             - adding a "KeyboardNavigation"-property to let the user decide,which navigation-behaviour (suggested by Jean-Pierre Martel)

***ATTENTION!!!*** There is an intrinsic limitation for the Gigagrid.
                   Due to the way it handles the scrolling (windows-messages),
                   it is usable only for rowsets, with not more then
                   32768/ height (in pixel) of one gridcell.

Thanks to Jean-Pierre Martel and Jrgen Feder, for countless advices and the help
in many questions.
*/

class GigaGrid(parentObj, name) of CONTAINER(parentObj, name) custom
   with (this)
      onOpen = class::GigaGrid_ONOPEN
      onDesignOpen = class::GigaGrid_ONDESIGNOPEN
      metric = 6	// Pixels
      borderStyle = 2	// Lowered
   endwith

   Form.metric = 6
   this.width = 180
   this.height = 180
   this.offset = (PUSHBUTTON_HEIGHT*3) + 2 // if the relative position =0 this has to be added to get the real top of the elevator
   this.roffset = 0
   this.rowheight= 0
   this.nTopDisplayedRow = 1 // nr. of the topmost-line in the grid (represents either record-number or sequence-number)
   this.nLastTopDisplayedRow = 0 // is the maximum-nr, that is allowed for topline (to avoid the grey-zone)
   this.selfcount = false	// if false, the user holds track of the nr. of rows, if true, the grid does the count(much slower)
   this.nRows= 0 // holds the nr of rows (count())
   this.refreshWhileScrolling = true // if you don't want the smooth scrolling, set this to false
   this.proportional = false // determines the way, clicks in the shaft are handled.
   								// false --> move elevator for one pixel
                           // true--> move the elevator to the clicked position
	this.highlightrow = true // is used in the same way as rowselect
   this.selectedbackgroundcolor = 0xdda000 // specifies the color for the highlighted row
   this.selectedforegroundcolor = 0x000000 // color of the font of the highlighted row
   this.hdc = 0       // to hold the devicecontext of the grid
	this.navigationtype = 1 //1=next,2=first,3=last,4=other (click or seeker)
	this.nrowsdisplacement = 0 // if navigationtype=1 this is the displacement <0 or >0
  	this.elevcol = ELEVATOR_COLOR
   this.norowset = false	// to prevent errors, when the developper forgets to assign a rowset to the datalink of the grid

   this.GRID1 = new GRID(this)
   with (this.GRID1)
   	onopen=class::Grid_onOpen
      bgColor = "white"
      cellHeight = 19
      hasIndicator = true				// important, must be set to true
      hasColumnHeadings = false
      hasVScrollHintText = false
      allowrowsizing=false
      rowSelect = false
      hScrollBar = 0	// Off
      vScrollBar = 0	// Off
      height = this.height
      left = 0
      top = 0
      allowediting=false
      width = this.width
      lockedcolumns = columncount   
   endwith

   this.PAINTBOX1 = new PAINTBOX(this) // used to surface hidden grid-actions, like
   with(this.PAINTBOX1)						//	the interaction with the wheel of the mouse
   	top=0
      left=0
      width=0
      height=0
      onLeftMouseDown = class::PAINTBOX_ONLEFTMOUSEDOWN
      onLeftMouseUp = class::PAINTBOX_ONLEFTMOUSEUP
   endwith

   this.SHOWCONTAINER = new CONTAINER(this)
   with (this.SHOWCONTAINER)
      width = this.width - 40 - this.grid1.gridlinewidth
      left = 22 + this.grid1.gridlinewidth
      top = 1
      height = this.grid1.cellheight
      borderStyle = 3	// none
      colornormal="0xdda000"
      visible = false
   endwith

   this.SCROLLCONTAINER = new CONTAINER(this)
   with (this.SCROLLCONTAINER)
      onOpen = class::SCROLLCONTAINER_ONOPEN
      width = PUSHBUTTON_WIDTH + 3
      left = this.width - PUSHBUTTON_WIDTH - 5
      top = 1
      height = this.height - 1 - HAS_HSCROLLBAR
      borderStyle = 9	// Etched In
   endwith

   this.SCROLLCONTAINER.SHAFT = new RECTANGLE(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.SHAFT)
      onLeftMouseDown = class::SHAFT_ONLEFTMOUSEDOWN
      onLeftMouseUp = class::SHAFT_ONLEFTMOUSEUP
      onmousemove = class::SHAFT_ONMOUSEMOVE
      left = 0
      top = (PUSHBUTTON_HEIGHT*3) + 1
      width = PUSHBUTTON_WIDTH - 1
      height = this.ScrollContainer.height - (PUSHBUTTON_HEIGHT*6) - 5
      text = ""
      border = false
      borderStyle = 2	// Lowered
   endwith

   this.scrollcontainer.shaft.pushed = false
   this.scrollcontainer.shaft.offset = 0   // vertical position of the click in the elevator
   this.scrollcontainer.shaft.ndelta = 0   // used for the refreshWhileScrolling-behaviour

   this.SCROLLCONTAINER.ELEVATOR = new IMAGE(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.ELEVATOR)
      onLeftMouseDown = class::ELEVATOR_ONLEFTMOUSEDOWN
      mousePointer = 13 // Hand
      height = 28
      left = 1
      top = (PUSHBUTTON_HEIGHT*3) + 2
      width = PUSHBUTTON_WIDTH - 3  // 16 = true width
      dataSource = 'RESOURCE E_' + ELEVATOR_COLOR + ' "VSB.dll"'
      borderStyle = 3	// None
   endwith

   this.SCROLLCONTAINER.PUSHBUTTON1 = new ScrollBUTTON(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.PUSHBUTTON1)
      onLeftMouseUp = class::HOME_ONCLICK
      top = 0
      upBitmap = 'RESOURCE Home_' + UPBITMAP_COLOR + ' "VSB.dll"'
      downBitmap = 'RESOURCE Home_Down "VSB.dll"'
      disabledBitmap = 'RESOURCE Home_Down "VSB.dll"'
      enabled=true
   endwith

   this.SCROLLCONTAINER.PUSHBUTTON2 = new scrollBUTTON(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.PUSHBUTTON2)
      onLeftMouseUp = class::PGUP_ONCLICK
      top = PUSHBUTTON_HEIGHT
      upBitmap = 'RESOURCE PgUp_' + UPBITMAP_COLOR + ' "VSB.dll"'
      downBitmap = 'RESOURCE PgUp_Down "VSB.dll"'
      disabledBitmap = 'RESOURCE PgUp_Down "VSB.dll"'
   endwith

   this.SCROLLCONTAINER.PUSHBUTTON3 = new ScrollBUTTON(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.PUSHBUTTON3)
      onLeftMouseDown = class::PREVIOUS_onLeftMouseDown
      top = PUSHBUTTON_HEIGHT*2
      upBitmap = 'RESOURCE Previous_' + UPBITMAP_COLOR + ' "VSB.dll"'
      downBitmap = 'RESOURCE Previous_Down "VSB.dll"'
      disabledBitmap = 'RESOURCE Previous_Down "VSB.dll"'
   endwith

   this.SCROLLCONTAINER.PUSHBUTTON4 = new scrollbutton(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.PUSHBUTTON4)
      onLeftMouseDown = class::NEXT_onLeftMouseDown
      top = this.ScrollContainer.height - (PUSHBUTTON_HEIGHT*3) - 3
      upBitmap = 'RESOURCE Next_' + UPBITMAP_COLOR + ' "VSB.dll"'
      downBitmap = 'RESOURCE Next_Down "VSB.dll"'
      disabledBitmap = 'RESOURCE Next_Down "VSB.dll"'
   endwith

   this.SCROLLCONTAINER.PUSHBUTTON5 = new ScrollBUTTON(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.PUSHBUTTON5)
      onLeftMouseUp = class::PGDOWN_ONCLICK
      top = this.ScrollContainer.height - (PUSHBUTTON_HEIGHT*2) - 3
      upBitmap = 'RESOURCE PgDown_' + UPBITMAP_COLOR + ' "VSB.dll"'
      downBitmap = 'RESOURCE PgDown_Down "VSB.dll"'
      disabledBitmap = 'RESOURCE PgDown_Down "VSB.dll"'
   endwith

   this.SCROLLCONTAINER.PUSHBUTTON6 = new ScrollBUTTON(this.SCROLLCONTAINER)
   with (this.SCROLLCONTAINER.PUSHBUTTON6)
      onLeftMouseUp = class::END_ONCLICK
      top = this.ScrollContainer.height - PUSHBUTTON_HEIGHT - 3
      upBitmap = 'RESOURCE End_' + UPBITMAP_COLOR + ' "VSB.dll"'
      downBitmap = 'RESOURCE End_Down "VSB.dll"'
      disabledBitmap = 'RESOURCE End_Down "VSB.dll"'
      enabled=true
   endwith

   Function GigaGrid_onDesignOpen
      local cShow, cScroll, PB1, PB2, PB3
      PB1 = PUSHBUTTON_HEIGHT     + 1
      PB2 = PUSHBUTTON_HEIGHT * 2 + 1
      PB3 = PUSHBUTTON_HEIGHT * 3 + 1
      cShow   = this.ShowContainer
      cScroll = this.ScrollContainer
      this.grid1.width = this.width

      cShow.width  = this.width - this.grid1.gridlinewidth - 40
      cShow.left   = this.grid1.gridlinewidth + 22
      cShow.top    = 1
      cShow.height = this.grid1.cellheight

      cScroll.width  = PUSHBUTTON_WIDTH + 3
      cScroll.left   = this.width - PUSHBUTTON_WIDTH - 5
      cScroll.top    = 1
      cScroll.height = this.height - 1 - HAS_HSCROLLBAR

      cScroll.shaft.left   = 0
      cScroll.shaft.top    = PB1
      cScroll.shaft.width  = PUSHBUTTON_WIDTH - 1
      cScroll.shaft.height = this.GRID1.height - PUSHBUTTON_HEIGHT*6 - 5

      cScroll.pushbutton1.top = 0
      cScroll.pushbutton2.top = 18
      cScroll.pushbutton3.top = 36
      cScroll.pushbutton4.top = cScroll.height - PB3
      cScroll.pushbutton5.top = cScroll.height - PB2
      cScroll.pushbutton6.top = cScroll.height - PB1

      cScroll.elevator.height = 28
      cScroll.elevator.left   = 1
      cScroll.elevator.top    = PB1 + 1
      cScroll.elevator.width  = PUSHBUTTON_WIDTH - 3  // 16 = true width
      return

	Function Init
   	// Definition of the needed Api-Calls 

      load dll "wndnew.dll"		// generates the connection between grid and paintbox
      extern cLONG    Adroffunc(cuint) wndnew
      extern cLONG    Wopos(clong,clong,clong,clong) wndnew

      * the following two calls are used to ensure, that once the user clicked the elevator,
      * all mouse-messages will get to him (even if the mouse-pointer leaves the elevator)
      extern chandle  SetCapture(chandle) user32
      extern clogical ReleaseCapture(cvoid) user32  

      * Used to force the grid too redraw, without moving the current row.
      extern cuint SendMessage(chandle,cuint,clong,clong) user32 from "SendMessageA"

      * To get and set different styles and procedures
      extern cuint    GetWindowLong(chandle,cint) user32 from "GetWindowLongA"
      extern clong    SetWindowLong(chandle,cint,clong) user32 from "SetWindowLongA"
      extern CHANDLE  GetWindow(chandle,cUINT) user32 
      extern clogical SetPropA(chandle, cptr, CHANDLE) user32 from "SetPropA"
      extern CHANDLE  GetProp(Chandle, cptr) user32 from "GetPropA"
      extern clogical SetWindowPos(chandle,chandle,cint,cint,cint,cint,cuint) user32

      * To avoid flickering of the grid in some special cases
      extern clogical LockWindowUpdate(chandle) user32
      extern clogical UpdateWindow(chandle) user32
      extern chandle  GetDesktopWindow(cvoid) user32

      * To handle GDI-Drawing
      extern chandle   GetDC(chandle) user32 
      return 

   Function GigaGrid_onOpen
      // the following lines to fill the container (or eventually the custom class object)
      // if it was resized in the Form designer or lowered by a menu
      class::init()
      this.borderstyle       := 3 // none
      this.Grid1.top         := 0
      this.Grid1.left        := 0
      this.nTopDisplayedRow	= 1
      try
	   	this.grid1.datalink.ref = this
	      this.grid1.datalink.first()
   	   this.Grid1.datalink.onnavigate={|t,n| ;this.ref.navigationtype=t ; this.ref.nrowsdisplacement=n}
      catch(exception e)
      	this.scrollcontainer.visible = false
         this.norowset=true
      endtry
      /* the following replaces the windprocedure of the indicator bar of the grid
      so that all wm-paint messages are deviated to the paintbox.
      Each WM-Paint message means a changing in the grid, which can now be monitored
      via the paintbox-event
      */
      tmphandle=getwindow(this.grid1.hwnd,GW_CHILD)  // gets the handle of the first child of the grid
      tmphandle=getwindow(tmphandle,GW_HWNDNEXT)     // next child
      this.pabohandle=getwindow(tmphandle,GW_HWNDNEXT)     // third child in the grid is used for wm_paint message checking
      this.oldfunc=getwindowlong(this.pabohandle,GWL_WNDPROC)     //adress of the Window-Procedure
      setpropa(this.pabohandle,"O",this.oldfunc)
      setpropa(this.pabohandle,"A",this.Paintbox1.hwnd)
      newfunc=adroffunc(1)
      setwindowlong(this.pabohandle,GWL_WNDPROC,newfunc)

      /* Another change of wndprocedure, this time of the grid itself, to be able
      to use the wheelmouse on the Gigagrid, again wheel-messages are sent to the
      paintbox, where they are checked and transfered to real movement in the grid
      */
      this.oldfunc=getwindowlong(this.grid1.hwnd,GWL_WNDPROC)     //adress of the Window-Procedure
      setpropa(this.grid1.hwnd,"O",this.oldfunc)
      setpropa(this.grid1.hwnd,"A",this.Paintbox1.hwnd)
      newfunc=adroffunc(2)
      setwindowlong(this.grid1.hwnd,GWL_WNDPROC,newfunc)

      this.hdc=getdc(this.grid1.hwnd)
      this.grid1.onopen()
  		this.set_custom_properties()   
   return

	Function doCount()
   	if this.norowset
      	this.nrows = 0
      	return
      endif
      this.nrows = this.grid1.datalink.count()
      this.nLastTopDisplayedRow = this.nrows - this.pagecount
      this.scrollcontainer.elevator.visible = (this.nLastTopDisplayedRow>0)
      return 

   Function ScrollContainer_onOpen
      local height_3PB
      height_3PB = pushbutton_height*3
      this.width              = this.pushbutton1.width + 3
      this.pushbutton1.top    = 0
      this.pushbutton2.width  = this.pushbutton1.width
      this.pushbutton2.top    = this.pushbutton1.height
      this.pushbutton2.height = this.pushbutton1.height
      this.pushbutton3.width  = this.pushbutton1.width
      this.pushbutton3.top    = this.pushbutton1.height*2
      this.pushbutton3.height = this.pushbutton1.height
      this.pushbutton4.width  = this.pushbutton1.width
      this.pushbutton4.height = this.pushbutton1.height
      this.pushbutton5.width  = this.pushbutton1.width
      this.pushbutton5.height = this.pushbutton1.height
      this.pushbutton6.width  = this.pushbutton1.width
      this.pushbutton6.height = this.pushbutton1.height
      this.SHAFT.width        = this.pushbutton1.width - 1
      this.SHAFT.top          = height_3PB + 1
      this.elevator.width     = this.pushbutton1.width - 3
      this.elevator.top       = this.SHAFT.top + 1
      return

   Function PgUp_onClick
 		local giga, rowsBefore, i
      giga = this.parent.parent
      rowsBefore = giga.nTopDisplayedRow - giga.pagecount
   	if rowsBefore >= 0  // enough space to do a pageup
	      sendmessage(this.parent.parent.grid1.hwnd,wm_vscroll,sb_pageup,0)    // forces grid one page up
   		giga.nTopDisplayedRow -= giga.pagecount
      else
  		   for i = 1 to giga.nTopDisplayedRow - 1
			   sendmessage(giga.grid1.hwnd,wm_vscroll,sb_lineup,0)
		   endfor
			giga.nTopDisplayedRow = 1      	
      endif
      giga.adjustelevator()  // updating the position of the elevator
   return

   Function Previous_onLeftMouseDown
		local giga
      giga = this.parent.parent
   	if giga.nTopDisplayedRow > 1
	      sendmessage(giga.grid1.hwnd,wm_vscroll,sb_lineup,0)    // forces grid one line up
   		giga.nTopDisplayedRow --
         giga.adjustelevator()  // updating the position of the elevator
      endif
      return

   Function Shaft_onMouseMove(flags, col, row)
      local e, giga, tester, nFirst
      e = this.parent.elevator
      giga = this.parent.parent
      nFirst = giga.nTopDisplayedRow
      tester = row - this.offset + 1
      if this.pushed 
   	   if tester > 1 and tester <= giga.shaftheight + 1
			   e.top = row - this.offset + giga.offset
            giga.nTopDisplayedRow = ceiling((e.top-giga.offset)/giga.shaftheight*giga.nLastTopDisplayedRow)+1
            nFirst -= giga.nTopDisplayedRow
            if giga.refreshWhileScrolling
	            giga.adjustgrid(-nFirst)
            else
               this.nDelta += nFirst
            endif
         else
            if tester <=1  
               class::Home_onClick()
               this.offset = max(this.offset + tester, 1)
               return
            else
               this.offset = min(e.height,this.offset + tester - giga.shaftheight)
            endif			
         endif
      endif
      return

   Function Shaft_onLeftMouseDown(flags, col, row)
	   local e, giga, nFirst, tmp
	   e = this.parent.elevator
      giga = this.parent.parent
      if giga.proportional
      	if row < e.height + 1
         	class::Home_onClick()
            e.onLeftMouseDown(0, 0, row - e.top + giga.offset)
            return
         endif
         if row > giga.shaftheight
         	class::End_onClick()
            e.onLeftMouseDown(0, 0, row - e.top + giga.offset)
            return
         endif
		   nFirst = giga.nTopDisplayedRow
         tmp = row - int(e.height/2) + giga.offset
         giga.nTopDisplayedRow = ceiling((tmp-giga.offset)/giga.shaftheight*giga.nLastTopDisplayedRow)+1
         nFirst -= giga.nTopDisplayedRow
         giga.adjustgrid(-nFirst)
         giga.adjustelevator()
         e.onLeftMouseDown(0, 0, row - e.top + giga.offset)
      else
         if row < e.top - giga.offset
         	class::PgUp_onClick()
         else
         	class::PgDown_onClick()
         endif
      endif
      return

   Function Shaft_onLeftMouseUp
   	local giga
      giga = this.parent.parent
      try
	      this.parent.elevator.datasource = "Resource E_" + giga.elevcol + ' "vsb.dll"'
	      catch(exception e)
      endtry
   	releasecapture()
      this.pushed = false
      if not giga.refreshWhileScrolling
	      giga.adjustgrid(-this.ndelta)
  		   this.ndelta = 0
      endif
     return

   Function AdjustGrid(param1)
   	local g, h, gh, gw, j
      if param1 = 0
      	return
      endif
      g = this.grid1
      this.showcontainer.visible = false
  		gh = g.height
 	   gw = g.width
		j = ( (abs(param1) + 1) + iif(g.hascolumnheadings, 1, 0) ) * g.parent.rowheight // +2
      SetWindowpos(g.hwnd,0,0,0,gw,j,30)
 	   sendmessage(g.hwnd,wm_vscroll,iif(param1<=0,sb_pageup,sb_pagedown),0)
     	SetWindowPos(g.hwnd,0,0,0,gw,gh,30)
      updatewindow(this.hwnd)
      return

   Function Next_onLeftMouseDown
		local giga
      giga = this.parent.parent
   	if giga.nTopDisplayedRow <= giga.nLastTopDisplayedRow  // to avoid the gray-zone in the grid
	      sendmessage(giga.grid1.hwnd,wm_vscroll,sb_linedown,0)    // forces grid one line down
   		giga.nTopDisplayedRow ++
         giga.adjustelevator()  // updating the position of the elevator
      endif
      return

   Function PgDown_onClick
 		local giga, leftToScroll, i
      giga = this.parent.parent
      leftToScroll = giga.nLastTopDisplayedRow - giga.nTopDisplayedRow + 1
   	if leftToScroll > giga.pagecount  // to avoid the gray-zone in the grid
	      sendmessage(giga.grid1.hwnd,wm_vscroll,sb_pagedown,0)  // forces grid one page down
   		giga.nTopDisplayedRow += giga.pagecount
      else
  		   for i = 1 to leftToScroll
			   sendmessage(giga.grid1.hwnd,wm_vscroll,sb_linedown,0)
		   endfor
         giga.nTopDisplayedRow += leftToScroll
      endif
      giga.adjustelevator()  // updating the position of the elevator
	return

	Function Paintbox_onLeftMouseDown(keys,col,row)    // this is used to implement the wheelmouse
      local s
      s = this.parent.scrollcontainer
  	   if not empty(row)
  	   * if row # 0
         s.pushbutton4.onLeftMouseDown()
      else
         s.pushbutton3.onLeftMouseDown()
      endif
      return

   Function Paintbox_onLeftMouseUp
   	local giga, r, wop, a1, a2, a3, a4, ncount
      giga = this.parent
      if giga.norowset
      	return
      endif
      r = giga.grid1.datalink
      r.onNavigate = Null
      wop = wopos(giga.pagecount + 1, giga.rowheight, giga.hdc, giga.roffset)  // -2 if highlighted row can not be seen

      // to workaround the bug with disappearing-gridlines when grid has no heading
      // also used to implement the color of the highlighted row

      if giga.highlightrow  
			giga.showcontainer.top = giga.rowheight*(wop - iif(giga.grid1.hascolumnheadings,0,1)) + giga.rowheight + giga.grid1.gridlinewidth + 2
         giga.showcontainer.visible = iif(wop <- 1, false, true)
      endif

      // the following is used to implement the seeker-control
      if giga.navigationtype # 0
			if wop >= 0
      	r.notifycontrols = false
	      a1 = r.bookmark()
	      ncount = int(giga.nrows/2)
	      a2 = ncount
	      a3 = ncount
	      r.first()
	      do while ncount > 0 and a3 # 0
	         r.next(a3)
	         ncount = int(ncount/2 + .5)
	         a4 = r.bookmark()
            a3 = iif(a1 < a4, -ncount, iif(a1 > a4, ncount, 0))
            a2 += a3
	      enddo
      	r.goto(a1)
         giga.nTopDisplayedRow = a2 + 1 - wop
         r.notifycontrols = true
         giga.adjustelevator()
         endif
		endif

      // workaround for a glitch, if the user clicks the <Last-Button> in the Toolbar
      if giga.nTopDisplayedRow > giga.nLastTopDisplayedRow + 1
        giga.adjustgrid(giga.nLastTopDisplayedRow-giga.nTopDisplayedRow + 1)
        giga.nTopDisplayedRow = giga.nLastTopDisplayedRow + 1
        giga.adjustelevator()
      endif
      this.parent.navigationtype = 0
      r.onnavigate = {|t,n| ;this.ref.navigationtype=t ; this.ref.nrowsdisplacement=n;this.ref.flag1=true}
   return

   function Home_onClick
      local giga
      giga = this.parent.parent
      giga.adjustgrid(1 - giga.nTopDisplayedRow)
      giga.nTopDisplayedRow = 1
      giga.adjustelevator()
   return

   function End_onClick
   	local giga
      giga = this.parent.parent
      giga.adjustgrid(giga.nLastTopDisplayedRow + 1 - giga.nTopDisplayedRow)
      giga.nTopDisplayedRow = giga.nLastTopDisplayedRow + 1
      giga.adjustelevator()
   return   	

	function AdjustElevator
	   local e	
   	e = this.scrollcontainer.elevator
	   if this.selfcount
   		this.docount()  // slows the display down !
	   endif
   	e.top = int(this.shaftheight/this.nLastTopDisplayedRow*(this.nTopDisplayedRow-1))+this.offset
      return

   Function Elevator_onLeftMouseDown(a,b,row)
   	local s
      s = this.parent.shaft
      setcapture(s.hwnd)
      s.pushed = true
      s.offset = row
      try
        this.datasource = "Resource E_" + this.parent.parent.elevcol + '_D "vsb.dll"'
	      catch(exception e)
      endtry
      return   	

   Function Set_Custom_Properties
      // this function belongs to the custom class' container
      // it is used to initialize the visual aspect of the gigagrid
      // can be caled in a resize-event f.i.
      local g, s, height_3PB, i
      height_3PB = pushbutton_height*3
      g = this.grid1
      s = this.ScrollContainer
      this.rowheight		= g.cellheight+g.gridlinewidth // height of one cell including the border-line
      this.pagecount		= int(this.height/this.rowheight) // number of rows in the grid
      this.height			:=(this.pagecount+iif(g.hascolumnheadings,1,0))*this.rowheight+4 // force the grid to show only entire lines
      this.roffset		= int(this.rowheight/2)+iif(g.hasColumnHeadings,this.rowheight,0) // used for the rowselect = true
      g.height 			:=this.height
      g.width 				:=this.width
      s.left 				:=this.width - (s.width )
      s.height 			:=this.height - 1 - HAS_HSCROLLBAR
      s.pushbutton4.top :=s.height - height_3PB - 3
      s.pushbutton5.top :=s.height - (s.pushbutton1.height*2) - 3
      s.pushbutton6.top :=s.height - s.pushbutton1.height - 3
      s.shaft.height 	:=s.height - (height_3PB*2) - 5
      this.elevcol		:=s.elevator.datasource.substring(AT("_",s.elevator.datasource,1),RAT(' ',s.elevator.datasource,1))
      this.elevcol = ltrim(rtrim(this.elevcol))
      try
	      s.elevator.dataSource = 'RESOURCE E_' + this.elevcol + ' "VSB.dll"'
      catch(exception e)
      endtry
      this.shaftheight	= s.shaft.height-s.elevator.height-2
      this.docount()
      this.adjustelevator()
		for i=1 to g.columns.size
      	this.gpef[i].colornormal="0x"+itoh(this.selectedforegroundcolor)+"/0x"+itoh(this.selectedbackgroundcolor)
      endfor
	return

   Function Grid_onOpen
   	local zw1,i,i1,zw2,ef,li
      this.lockedcolumns=this.columncount
      zw1=this.columns.size
      if this.parent.norowset  
      	return
      endif
      this.parent.ShowContainer.visible = not (this.allowEditing = true)
      if zw1=0
      	zw1=this.datalink.fields.size
         this.columns=new array(zw1)
         zw2=0
         for i=1 to zw1
         	i1="column"+ltrim(rtrim(str(i)))
         	this.columns[i1]=new gridcolumn(this)
            this.columns[i1].datalink=this.datalink.fields[i]
            this.columns[i1].editortype=1
            this.columns[i1].width=this.width/zw1
            zw2 += this.columns[i1].width
         endfor
         this.parent.showcontainer.width = zw2
      endif
     	this.parent.gpef=new array(zw1+1)
      this.parent.linie=new array(zw1+1)
      zw2=0
		for i=1 to zw1
        	ef=this.parent.gpef
         li=this.parent.linie
         do case
            case this.columns[i].editortype=1
	            ef[i] = new entryfield(this.parent.showcontainer)
               ef[i].fontSize = iif(empty(this.parent.grid1.columns[i].editorcontrol.fontSize), this.parent.grid1.fontSize, this.parent.grid1.columns[i].editorcontrol.fontSize)
            case this.columns[i].editortype=2
            	ef[i] = new checkbox(this.parent.showcontainer)
               ef[i].text=''
            case this.columns[i].editortype=4
               ef[i] = new combobox(this.parent.showcontainer)
            otherwise
               ef[i] = new entryfield(this.parent.showcontainer)
               ef[i].fontSize = iif(empty(this.parent.grid1.columns[i].editorcontrol.fontSize), this.parent.grid1.fontSize, this.parent.grid1.columns[i].editorcontrol.fontSize)
         endcase
			ef[i].width = this.columns[i].width
         ef[i].left = zw2
         ef[i].datalink = this.columns[i].datalink
         if type("this.columns[i].editorcontrol.fontname")#"U"
				if type("this.fontsize")#"N"
	   	      * ef[i].fontsize = this.columns[i].editorcontrol.fontsize
               ef[i].fontname = this.columns[i].editorcontrol.fontname
            else
            	fs=this.columns[i].editorcontrol.fontsize
            	* ef[i].fontsize = iif(fs=0,this.fontsize,fs)
               fn=this.columns[i].editorcontrol.fontname
               ef[i].fontname = iif(fn='',this.fontname,fn)
            endif
         endif
         ef[i].colornormal = "0x"+itoh(this.parent.selectedforegroundcolor)+"/0x"+itoh(this.parent.selectedbackgroundcolor)
         ef[i].height = this.cellheight
  			this.parent.gpef[i].borderstyle=3        //  question of taste
         ef[i].when={; return false}
         zw2+= ef[i].width
			li[i] = new line(this.parent.showcontainer)
         li[i].top = 0
         li[i].left = zw2
         li[i].right = zw2
         li[i].bottom = this.cellheight
         li[i].width = this.gridlinewidth
         li[i].colornormal = "BtnFace"
         zw2+=this.gridlinewidth
      endfor
      this.parent.showcontainer.width=zw2
      this.parent.showcontainer.height = this.parent.grid1.cellHeight
 	return

endclass

class scrollbutton(parentobj) of pushbutton(parentobj)

with(this)
	left		:=0
   width		:= PUSHBUTTON_WIDTH
	height 	:= PUSHBUTTON_HEIGHT
   text 		:= ""
   speedBar := true
   borderStyle := 3	
endwith

endclass


